/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * From: @(#)announce.c	5.9 (Berkeley) 2/26/91
 */
char ann_rcsid[] = 
  "$Id: announce.c,v 1.4 1996/12/29 17:13:53 dholland Exp $";

#include "config.h"
#include <sys/stat.h>
#include <sys/ioctl.h>
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(s) ((unsigned)(s) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(s) (((s) & 0377) == 0)
#endif
#include <errno.h>
#include <signal.h>
#include <setjmp.h>
#include "proto.h"

static int announce_proc(CTL_MSG *request, const char *remote_machine, time_t ut_time);
static void print_mesg(FILE *tf, CTL_MSG *request, const char *remote_machine,
	int silent);

extern char hostname[];

/*
 * Announce an invitation to talk.
 *
 * Because the tty driver insists on attaching a terminal-less
 * process to any terminal that it writes on, we must fork a child
 * to protect ourselves
 */
int
announce(CTL_MSG *request, const char *remote_machine, time_t ut_time)
{
	int pid, val, status;

	if ((pid = fork())!=0) {
		/* we are the parent, so wait for the child */
		if (pid == -1)		/* the fork failed */
			return FAILED;
		do {
			val = wait(&status);
			if (val == -1) {
				if (errno == EINTR)
					continue;
				/* shouldn't happen */
				syslog(LOG_WARNING, "announce: wait: %m");
				return FAILED;
			}
		} while (val != pid);
		if (!WIFEXITED(status))	/* we were killed by some signal */
			return FAILED;
		/* Get the second byte, this is the exit/return code */
		return WEXITSTATUS(status);
	}
	/* we are the child, go and do it */
	_exit(announce_proc(request, remote_machine, ut_time));
}


/*
 * Open the tty.  It's possible that the /etc/utmp entry incorrectly
 * corresponds to an unused pseudo-tty, in which case the attempt to open it
 * will block until someone options the control side of the pseudo-tty,
 * which won't happen because telnetd sees someone has the slave side open.
 * So this just puts a ten second timeout on the open, and returns NULL if
 * the open fails.   -jdw
 */

jmp_buf jenv;

void ot_timeout() { longjmp(jenv,1); }

FILE *opentty(char *full_tty)
{
RETSIGTYPE (*oldalrm)();
FILE *tf;

	oldalrm= signal(SIGALRM,ot_timeout);
	alarm(10);
	if (!setjmp(jenv))
	{
		tf= fopen(full_tty, "w");
	}
	else
	{
		tf= NULL;	/* timed out */
	}
	signal(SIGALRM,oldalrm);
	return tf;
}

/*
 * See if the user is accepting messages. If so, announce that 
 * a talk is requested.
 */
static int
announce_proc(CTL_MSG *request, const char *remote_machine, time_t ut_time)
{
/*	int pid, status; */
	char full_tty[64];
	FILE *tf;
	struct stat stbuf;
#ifdef ORVILLE
	int rc, silent_flag;
#else
#define silent_flag 0
#endif

#ifdef HAVE_SNPRINTF
	snprintf(full_tty, sizeof(full_tty), "%s/%s", _PATH_DEV, 
		 request->r_tty);
#else
	sprintf(full_tty, "%.30s/%.30s", _PATH_DEV, request->r_tty);
#endif
	if (access(full_tty, 0) != 0)
		return FAILED;
	if ((tf = opentty(full_tty)) == NULL)
		return PERMISSION_DENIED;
	/*
	 * On first tty open, the server will have
	 * it's pgrp set, so disconnect us from the
	 * tty before we catch a signal.
	 */
	ioctl(fileno(tf), TIOCNOTTY, NULL);
	if (fstat(fileno(tf), &stbuf) < 0)
	{
		fclose(tf);
		return (PERMISSION_DENIED);
	}
#ifdef ORVILLE
	rc= maytalk(request->r_tty, ut_time, &stbuf,
		remote_machine, request->l_name, &silent_flag);
        if (rc != SUCCESS)
	{
		fclose(tf);
		return rc;
	}
#else
	if ((stbuf.st_mode&020) == 0)
	{
		fclose(tf);
		return PERMISSION_DENIED;
	}
#endif
	print_mesg(tf, request, remote_machine, silent_flag);
	fclose(tf);
	return SUCCESS;
}

/*
 * Reject control codes and other crap.
 * FUTURE: don't clear high ascii if the tty is in 8-bit mode
 * Note that we don't need to let through tabs or newlines here.
 */
static int safechar(int ch) {
    if (ch>127 || ch<32) ch = '?';
    return ch;
}

#define max(a,b) ( (a) > (b) ? (a) : (b) )
#define N_LINES 5
#define N_CHARS 256

/*
 * Build a block of characters containing the message. 
 * It is sent blank filled and in a single block to
 * try to keep the message in one piece if the recipient
 * in in vi at the time
 */
static void
print_mesg(FILE *tf, CTL_MSG *request, const char *remote_machine, int silent)
{
	struct timeval clocc;
	struct tm *localclock;
	char line_buf[N_LINES][N_CHARS];
	int sizes[N_LINES];
	char big_buf[N_LINES*N_CHARS];
	char *bptr, *lptr;
	int i, j, max_size;
	time_t footime;

	i = 0;
	max_size = 0;
	footime = time(NULL);
	localclock = localtime(&footime);
	sprintf(line_buf[i], " ");
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
#ifdef HAVE_SNPRINTF
	snprintf(line_buf[i], N_CHARS, 
		 "Message from Talk_Daemon@%s at %d:%02d ...",
		 hostname, localclock->tm_hour, localclock->tm_min);
#else
	sprintf(line_buf[i], "Message from Talk_Daemon@%.80s at %d:%02d ...",
		 hostname, localclock->tm_hour, localclock->tm_min);
#endif
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
#ifdef HAVE_SNPRINTF
	snprintf(line_buf[i], N_CHARS, "talk: connection requested by %s@%s.",
		request->l_name, remote_machine);
#else
	sprintf(line_buf[i], "talk: connection requested by %.80s@%.80s.",
		request->l_name, remote_machine);
#endif
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
#ifdef HAVE_SNPRINTF
	snprintf(line_buf[i], N_CHARS, "talk: respond with:  talk %s@%s",
		 request->l_name, remote_machine);
#else
	sprintf(line_buf[i], "talk: respond with:  talk %.80s@%.80s",
		 request->l_name, remote_machine);
#endif
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	sprintf(line_buf[i], " ");
	sizes[i] = strlen(line_buf[i]);
	max_size = max(max_size, sizes[i]);
	i++;
	bptr = big_buf;
	if (!silent)
		*bptr++ = '\007'; /* send something to wake them up */
	*bptr++ = '\r';	/* add a \r in case of raw mode */
	*bptr++ = '\n';
	for (i = 0; i < N_LINES; i++) {
		/* copy the line into the big buffer */
		lptr = line_buf[i];
		while (*lptr != '\0')
			*(bptr++) = safechar(*(lptr++));
		/* pad out the rest of the lines with blanks */
		for (j = sizes[i]; j < max_size + 2; j++)
			*(bptr++) = ' ';
		*(bptr++) = '\r';	/* add a \r in case of raw mode */
		*(bptr++) = '\n';
	}
	*bptr = 0;
	fprintf(tf, big_buf);
	fflush(tf);
	ioctl(fileno(tf), TIOCNOTTY, NULL);
}
